'''tldextract GAE app setup.py'''

import os
import sys


def main():
    # Must include deps in this folder for GAE. It doesn't use a e.g.
    # requirements.txt.
    app_folder = os.path.dirname(__file__)
    sys.path.append(os.path.abspath(os.path.join(app_folder, os.pardir)))
    deps = ('tldextract', 'web')
    for modname in deps:
        symlink = os.path.join(app_folder, modname)
        try:
            os.remove(symlink)
        except (OSError, IOError):
            pass

        mod = __import__(modname)
        loc = os.path.dirname(mod.__file__)
        os.symlink(loc, symlink)


if __name__ == "__main__":
    main()



'tldextract helpers for testing and fetching remote resources.'


import logging
import re
import socket
import sys

import requests
from requests_file import FileAdapter

# pylint: disable=import-error,invalid-name,no-name-in-module,redefined-builtin
if sys.version_info < (3,):  # pragma: no cover
    from urlparse import scheme_chars
else:  # pragma: no cover
    from urllib.parse import scheme_chars
    unicode = str
# pylint: enable=import-error,invalid-name,no-name-in-module,redefined-builtin


IP_RE = re.compile(r'^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$')  # pylint: disable=line-too-long

PUNY_RE = re.compile(r'^xn--', re.IGNORECASE)

SCHEME_RE = re.compile(r'^([' + scheme_chars + ']+:)?//')

LOG = logging.getLogger('tldextract')


def find_first_response(urls, cache_fetch_timeout=None):
    """ Decode the first successfully fetched URL, from UTF-8 encoding to
    Python unicode.
    """
    with requests.Session() as session:
        session.mount('file://', FileAdapter())

        for url in urls:
            try:
                text = session.get(url, timeout=cache_fetch_timeout).text
            except requests.exceptions.RequestException:
                LOG.exception(
                    'Exception reading Public Suffix List url %s',
                    url
                )
            else:
                return _decode_utf8(text)

    LOG.error(
        'No Public Suffix List found. Consider using a mirror or constructing '
        'your TLDExtract with `suffix_list_urls=None`.'
    )
    return unicode('')


def _decode_utf8(text):
    """ Decode from utf8 to Python unicode string.

    The suffix list, wherever its origin, should be UTF-8 encoded.
    """
    if not isinstance(text, unicode):
        return unicode(text, 'utf-8')
    else:
        return text


def looks_like_ip(maybe_ip):
    """Does the given str look like an IP address?"""
    if not maybe_ip[0].isdigit():
        return False

    try:
        socket.inet_aton(maybe_ip)
        return True
    except (AttributeError, UnicodeError):
        if IP_RE.match(maybe_ip):
            return True
    except socket.error:
        return False



'''web.py handlers for making a JSON-over-HTTP API around tldextract.'''

import json

# pylint: disable=import-error
import tldextract
import web

URLS = (
    '/api/extract', 'Extract',
    '/api/re', 'TLDSet',
    '/test', 'Test',
)


class Extract(object):

    def GET(self):  # pylint: disable=invalid-name,no-self-use
        url = web.input(url='').url
        if not url:
            return web.webapi.badrequest()

        ext = tldextract.extract(url)._asdict()
        web.header('Content-Type', 'application/json')
        return json.dumps(ext) + '\n'


class TLDSet(object):

    def GET(self):  # pylint: disable=invalid-name,no-self-use
        web.header('Content-Type', 'text/html; charset=utf-8')
        return '<br/>'.join(sorted(tldextract.tldextract.TLD_EXTRACTOR.tlds))


APP = web.application(URLS, globals())
main = APP.cgirun()



# -*- coding: utf-8 -*-
'''Main tldextract unit tests.'''

import sys

import responses
import tldextract
from .helpers import temporary_file
if sys.version_info >= (3,):  # pragma: no cover
    unicode = str  # pylint: disable=invalid-name,redefined-builtin


# pylint: disable=invalid-name
extract = tldextract.TLDExtract(cache_file=temporary_file())
extract_no_cache = tldextract.TLDExtract(cache_file=False)
extract_using_real_local_suffix_list = tldextract.TLDExtract(cache_file=temporary_file())
extract_using_real_local_suffix_list_no_cache = tldextract.TLDExtract(cache_file=False)
extract_using_fallback_to_snapshot_no_cache = tldextract.TLDExtract(
    cache_file=None,
    suffix_list_urls=None
)
# pylint: enable=invalid-name


def assert_extract(
        url,
        expected_domain_data,
        expected_ip_data='',
        funs=(
            extract,
            extract_no_cache,
            extract_using_real_local_suffix_list,
            extract_using_real_local_suffix_list_no_cache,
            extract_using_fallback_to_snapshot_no_cache
        )):
    (expected_fqdn,
     expected_subdomain,
     expected_domain,
     expected_tld) = expected_domain_data
    for fun in funs:
        ext = fun(url)
        assert expected_fqdn == ext.fqdn
        assert expected_subdomain == ext.subdomain
        assert expected_domain == ext.domain
        assert expected_tld == ext.suffix
        assert expected_ip_data == ext.ipv4


def test_american():
    assert_extract('http://www.google.com',
                   ('www.google.com', 'www', 'google', 'com'))


def test_british():
    assert_extract("http://www.theregister.co.uk",
                   ("www.theregister.co.uk", "www", "theregister", "co.uk"))


def test_no_subdomain():
    assert_extract("http://gmail.com", ("gmail.com", "", "gmail", "com"))


def test_nested_subdomain():
    assert_extract("http://media.forums.theregister.co.uk",
                   ("media.forums.theregister.co.uk", "media.forums",
                    "theregister", "co.uk"))


def test_odd_but_possible():
    assert_extract('http://www.www.com', ('www.www.com', 'www', 'www', 'com'))
    assert_extract('http://www.com', ('www.com', '', 'www', 'com'))


def test_local_host():
    assert_extract('http://internalunlikelyhostname/',
                   ('', '', 'internalunlikelyhostname', ''))
    assert_extract('http://internalunlikelyhostname.bizarre',
                   ('', 'internalunlikelyhostname', 'bizarre', ''))


def test_qualified_local_host():
    assert_extract('http://internalunlikelyhostname.info/',
                   ('internalunlikelyhostname.info',
                    '', 'internalunlikelyhostname', 'info'))
    assert_extract('http://internalunlikelyhostname.information/',
                   ('',
                    'internalunlikelyhostname', 'information', ''))


def test_ip():
    assert_extract('http://216.22.0.192/',
                   ('', '', '216.22.0.192', ''),
                   expected_ip_data='216.22.0.192',)
    assert_extract('http://216.22.project.coop/',
                   ('216.22.project.coop', '216.22', 'project', 'coop'))


def test_looks_like_ip():
    assert_extract(u'1\xe9', ('', '', u'1\xe9', ''))


def test_punycode():
    assert_extract('http://xn--h1alffa9f.xn--p1ai',
                   ('xn--h1alffa9f.xn--p1ai', '', 'xn--h1alffa9f', 'xn--p1ai'))
    assert_extract('http://xN--h1alffa9f.xn--p1ai',
                   ('xN--h1alffa9f.xn--p1ai', '', 'xN--h1alffa9f', 'xn--p1ai'))
    assert_extract('http://XN--h1alffa9f.xn--p1ai',
                   ('XN--h1alffa9f.xn--p1ai', '', 'XN--h1alffa9f', 'xn--p1ai'))
    # Entries that might generate UnicodeError exception
    # This subdomain generates UnicodeError 'IDNA does not round-trip'
    assert_extract('xn--tub-1m9d15sfkkhsifsbqygyujjrw602gk4li5qqk98aca0w.google.com',
                   ('xn--tub-1m9d15sfkkhsifsbqygyujjrw602gk4li5qqk98aca0w.google.com',
                    'xn--tub-1m9d15sfkkhsifsbqygyujjrw602gk4li5qqk98aca0w', 'google',
                    'com'))
    # This subdomain generates UnicodeError 'incomplete punicode string'
    assert_extract('xn--tub-1m9d15sfkkhsifsbqygyujjrw60.google.com',
                   ('xn--tub-1m9d15sfkkhsifsbqygyujjrw60.google.com',
                    'xn--tub-1m9d15sfkkhsifsbqygyujjrw60', 'google', 'com'))


def test_invalid_puny_with_puny():
    assert_extract('http://xn--zckzap6140b352by.blog.so-net.xn--wcvs22d.hk',
                   ('xn--zckzap6140b352by.blog.so-net.xn--wcvs22d.hk',
                    'xn--zckzap6140b352by.blog', 'so-net', 'xn--wcvs22d.hk'))


def test_puny_with_non_puny():
    assert_extract(u'http://xn--zckzap6140b352by.blog.so-net.教育.hk',
                   (u'xn--zckzap6140b352by.blog.so-net.教育.hk',
                    'xn--zckzap6140b352by.blog', 'so-net', u'教育.hk'))


def test_idna_2008():
    """Python supports IDNA 2003.
    The IDNA library adds 2008 support for characters like ß.
    """
    assert_extract('xn--gieen46ers-73a.de',
                   ('xn--gieen46ers-73a.de', '', 'xn--gieen46ers-73a', 'de'))


def test_empty():
    assert_extract('http://', ('', '', '', ''))


def test_scheme():
    assert_extract('https://mail.google.com/mail', ('mail.google.com', 'mail', 'google', 'com'))
    assert_extract('ssh://mail.google.com/mail', ('mail.google.com', 'mail', 'google', 'com'))
    assert_extract('//mail.google.com/mail', ('mail.google.com', 'mail', 'google', 'com'))
    assert_extract('mail.google.com/mail',
                   ('mail.google.com', 'mail', 'google', 'com'), funs=(extract,))


def test_port():
    assert_extract('git+ssh://www.github.com:8443/', ('www.github.com', 'www', 'github', 'com'))


def test_username():
    assert_extract('ftp://johndoe:5cr1p7k1dd13@1337.warez.com:2501',
                   ('1337.warez.com', '1337', 'warez', 'com'))


def test_query_fragment():
    assert_extract('http://google.com?q=cats', ('google.com', '', 'google', 'com'))
    assert_extract('http://google.com#Welcome', ('google.com', '', 'google', 'com'))
    assert_extract('http://google.com/#Welcome', ('google.com', '', 'google', 'com'))
    assert_extract('http://google.com/s#Welcome', ('google.com', '', 'google', 'com'))
    assert_extract('http://google.com/s?q=cats#Welcome', ('google.com', '', 'google', 'com'))


def test_regex_order():
    assert_extract('http://www.parliament.uk',
                   ('www.parliament.uk', 'www', 'parliament', 'uk'))
    assert_extract('http://www.parliament.co.uk',
                   ('www.parliament.co.uk', 'www', 'parliament', 'co.uk'))


def test_unhandled_by_iana():
    assert_extract('http://www.cgs.act.edu.au/',
                   ('www.cgs.act.edu.au', 'www', 'cgs', 'act.edu.au'))
    assert_extract('http://www.google.com.au/',
                   ('www.google.com.au', 'www', 'google', 'com.au'))


def test_tld_is_a_website_too():
    assert_extract('http://www.metp.net.cn', ('www.metp.net.cn', 'www', 'metp', 'net.cn'))
    # This is unhandled by the PSL. Or is it?
    # assert_extract(http://www.net.cn',
    #                ('www.net.cn', 'www', 'net', 'cn'))


def test_dns_root_label():
    assert_extract('http://www.example.com./',
                   ('www.example.com', 'www', 'example', 'com'))


def test_private_domains():
    assert_extract('http://waiterrant.blogspot.com',
                   ('waiterrant.blogspot.com', 'waiterrant', 'blogspot', 'com'))


def test_ipv4():
    assert_extract('http://127.0.0.1/foo/bar',
                   ('', '', '127.0.0.1', ''),
                   expected_ip_data='127.0.0.1')


def test_ipv4_bad():
    assert_extract('http://256.256.256.256/foo/bar',
                   ('', '256.256.256', '256', ''),
                   expected_ip_data='')


def test_ipv4_lookalike():
    assert_extract('http://127.0.0.1.9/foo/bar',
                   ('', '127.0.0.1', '9', ''),
                   expected_ip_data='')


def test_result_as_dict():
    result = extract(
        "http://admin:password1@www.google.com:666"
        "/secret/admin/interface?param1=42"
    )
    expected_dict = {'subdomain': 'www',
                     'domain': 'google',
                     'suffix': 'com'}
    assert result._asdict() == expected_dict


@responses.activate  # pylint: disable=no-member
def test_cache_timeouts():
    server = 'http://some-server.com'
    responses.add(  # pylint: disable=no-member
        responses.GET,  # pylint: disable=no-member
        server,
        status=408
    )

    assert tldextract.remote.find_first_response([server], 5) == unicode('')



'''py.test standard config file.'''

import sys

# pylint: disable=invalid-name

collect_ignore = ('setup.py',)

# pylint: enable=invalid-name

def pytest_cmdline_preparse(args):
    is_pylint_compatible = sys.version_info < (3, 6)
    if not is_pylint_compatible:
        args.remove('--pylint')



'''tldextract CLI'''


import logging

try:
    import pkg_resources
    __version__ = pkg_resources.get_distribution('tldextract').version  # pylint: disable=no-member
except ImportError:
    __version__ = '(local)'
except pkg_resources.DistributionNotFound:
    __version__ = '(local)'

from .tldextract import TLDExtract

try:
    unicode
except NameError:
    unicode = str  # pylint: disable=invalid-name,redefined-builtin


def main():
    '''tldextract CLI main command.'''
    import argparse

    logging.basicConfig()

    parser = argparse.ArgumentParser(
        prog='tldextract',
        description='Parse hostname from a url or fqdn')

    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)  # pylint: disable=no-member
    parser.add_argument('input', metavar='fqdn|url',
                        type=unicode, nargs='*', help='fqdn or url')

    parser.add_argument('-u', '--update', default=False, action='store_true',
                        help='force fetch the latest TLD definitions')
    parser.add_argument('-c', '--cache_file',
                        help='use an alternate TLD definition file')
    parser.add_argument('-p', '--private_domains', default=False, action='store_true',
                        help='Include private domains')

    args = parser.parse_args()
    tld_extract = TLDExtract(include_psl_private_domains=args.private_domains)

    if args.cache_file:
        tld_extract.cache_file = args.cache_file

    if args.update:
        tld_extract.update(True)
    elif len(args.input) is 0:
        parser.print_usage()
        exit(1)

    for i in args.input:
        print(' '.join(tld_extract(i)))  # pylint: disable=superfluous-parens



"""`tldextract` accurately separates the gTLD or ccTLD (generic or country code
top-level domain) from the registered domain and subdomains of a URL.

    >>> import tldextract
    >>> tldextract.extract('http://forums.news.cnn.com/')
    ExtractResult(subdomain='forums.news', domain='cnn', suffix='com')
    >>> tldextract.extract('http://forums.bbc.co.uk/') # United Kingdom
    ExtractResult(subdomain='forums', domain='bbc', suffix='co.uk')
    >>> tldextract.extract('http://www.worldbank.org.kg/') # Kyrgyzstan
    ExtractResult(subdomain='www', domain='worldbank', suffix='org.kg')

`ExtractResult` is a namedtuple, so it's simple to access the parts you want.

    >>> ext = tldextract.extract('http://forums.bbc.co.uk')
    >>> (ext.subdomain, ext.domain, ext.suffix)
    ('forums', 'bbc', 'co.uk')
    >>> # rejoin subdomain and domain
    >>> '.'.join(ext[:2])
    'forums.bbc'
    >>> # a common alias
    >>> ext.registered_domain
    'bbc.co.uk'

By default, this package supports the public ICANN TLDs and their exceptions.
You can optionally support the Public Suffix List's private domains as well.
"""

import sys
from setuptools import setup

if sys.version_info < (2, 7):
    raise RuntimeError("Python 2.6 is EOL and no longer supported. "
                       "Please upgrade your Python or use an older "
                       "version of tldextract.")

INSTALL_REQUIRES = ["setuptools", "idna", "requests>=2.1.0", "requests-file>=1.4"]

setup(
    name="tldextract",
    version="2.2.1",
    author="John Kurkowski",
    author_email="john.kurkowski@gmail.com",
    description=("Accurately separate the TLD from the registered domain and "
                 "subdomains of a URL, using the Public Suffix List. By "
                 "default, this includes the public ICANN TLDs and their "
                 "exceptions. You can optionally support the Public Suffix "
                 "List's private domains as well."),
    license="BSD License",
    keywords="tld domain subdomain url parse extract urlparse urlsplit public suffix list",
    url="https://github.com/john-kurkowski/tldextract",
    packages=['tldextract'],
    include_package_data=True,
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
    long_description=__doc__,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    entry_points={
        'console_scripts': [
            'tldextract = tldextract.cli:main', ]
    },
    install_requires=INSTALL_REQUIRES,
)



'''tldextract unit tests with a custom suffix list.'''

import os

import tldextract

from .helpers import temporary_file


FAKE_SUFFIX_LIST_URL = "file://" + os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'fixtures/fake_suffix_list_fixture.dat'
)
EXTRA_SUFFIXES = ['foo1', 'bar1', 'baz1']

# pylint: disable=invalid-name
extract_using_fake_suffix_list = tldextract.TLDExtract(
    cache_file=temporary_file(),
    suffix_list_urls=[FAKE_SUFFIX_LIST_URL]
)
extract_using_fake_suffix_list_no_cache = tldextract.TLDExtract(
    cache_file=None,
    suffix_list_urls=[FAKE_SUFFIX_LIST_URL]
)
extract_using_extra_suffixes = tldextract.TLDExtract(
    cache_file=None,
    suffix_list_urls=[FAKE_SUFFIX_LIST_URL],
    extra_suffixes=EXTRA_SUFFIXES
)
# pylint: enable=invalid-name


def test_suffix_which_is_not_in_custom_list():
    for fun in (extract_using_fake_suffix_list, extract_using_fake_suffix_list_no_cache):
        result = fun("www.google.com")
        assert result.suffix == ""


def test_custom_suffixes():
    for fun in (extract_using_fake_suffix_list, extract_using_fake_suffix_list_no_cache):
        for custom_suffix in ('foo', 'bar', 'baz'):
            result = fun("www.foo.bar.baz.quux" + "." + custom_suffix)
            assert result.suffix == custom_suffix


def test_suffix_which_is_not_in_extra_list():
    result = extract_using_extra_suffixes("www.google.com")
    assert result.suffix == ""


def test_extra_suffixes():
    for custom_suffix in EXTRA_SUFFIXES:
        netloc = "www.foo.bar.baz.quux" + "." + custom_suffix
        result = extract_using_extra_suffixes(netloc)
        assert result.suffix == custom_suffix



# -*- coding: utf-8 -*-
"""`tldextract` accurately separates the gTLD or ccTLD (generic or country code
top-level domain) from the registered domain and subdomains of a URL.

    >>> import tldextract

    >>> tldextract.extract('http://forums.news.cnn.com/')
    ExtractResult(subdomain='forums.news', domain='cnn', suffix='com')

    >>> tldextract.extract('http://forums.bbc.co.uk/') # United Kingdom
    ExtractResult(subdomain='forums', domain='bbc', suffix='co.uk')

    >>> tldextract.extract('http://www.worldbank.org.kg/') # Kyrgyzstan
    ExtractResult(subdomain='www', domain='worldbank', suffix='org.kg')

`ExtractResult` is a namedtuple, so it's simple to access the parts you want.

    >>> ext = tldextract.extract('http://forums.bbc.co.uk')
    >>> (ext.subdomain, ext.domain, ext.suffix)
    ('forums', 'bbc', 'co.uk')
    >>> # rejoin subdomain and domain
    >>> '.'.join(ext[:2])
    'forums.bbc'
    >>> # a common alias
    >>> ext.registered_domain
    'bbc.co.uk'

Note subdomain and suffix are _optional_. Not all URL-like inputs have a
subdomain or a valid suffix.

    >>> tldextract.extract('google.com')
    ExtractResult(subdomain='', domain='google', suffix='com')

    >>> tldextract.extract('google.notavalidsuffix')
    ExtractResult(subdomain='google', domain='notavalidsuffix', suffix='')

    >>> tldextract.extract('http://127.0.0.1:8080/deployed/')
    ExtractResult(subdomain='', domain='127.0.0.1', suffix='')

If you want to rejoin the whole namedtuple, regardless of whether a subdomain
or suffix were found:

    >>> ext = tldextract.extract('http://127.0.0.1:8080/deployed/')
    >>> # this has unwanted dots
    >>> '.'.join(ext)
    '.127.0.0.1.'
    >>> # join part only if truthy
    >>> '.'.join(part for part in ext if part)
    '127.0.0.1'
"""


import collections
from contextlib import closing
import errno
from functools import wraps
import json
import logging
import os
import re

import idna

try:
    import pkg_resources
except ImportError:
    class pkg_resources(object):  # pylint: disable=invalid-name

        """Fake pkg_resources interface which falls back to getting resources
        inside `tldextract`'s directory.
        """
        @classmethod
        def resource_stream(cls, _, resource_name):
            moddir = os.path.dirname(__file__)
            path = os.path.join(moddir, resource_name)
            return open(path)

from .remote import find_first_response
from .remote import looks_like_ip
from .remote import IP_RE
from .remote import PUNY_RE
from .remote import SCHEME_RE

# pylint: disable=invalid-name,undefined-variable
try:
    STRING_TYPE = basestring
except NameError:
    STRING_TYPE = str
# pylint: enable=invalid-name,undefined-variable


LOG = logging.getLogger("tldextract")

CACHE_FILE_DEFAULT = os.path.join(os.path.dirname(__file__), '.tld_set')
CACHE_FILE = os.path.expanduser(os.environ.get("TLDEXTRACT_CACHE", CACHE_FILE_DEFAULT))
CACHE_TIMEOUT = os.environ.get('TLDEXTRACT_CACHE_TIMEOUT')

PUBLIC_SUFFIX_LIST_URLS = (
    'https://publicsuffix.org/list/public_suffix_list.dat',
    'https://raw.githubusercontent.com/publicsuffix/list/master/public_suffix_list.dat',
)

PUBLIC_SUFFIX_RE = re.compile(r'^(?P<suffix>[.*!]*\w[\S]*)', re.UNICODE | re.MULTILINE)


class ExtractResult(collections.namedtuple('ExtractResult', 'subdomain domain suffix')):
    '''namedtuple of a URL's subdomain, domain, and suffix.'''

    # Necessary for __dict__ member to get populated in Python 3+
    __slots__ = ()

    @property
    def registered_domain(self):
        """
        Joins the domain and suffix fields with a dot, if they're both set.

        >>> extract('http://forums.bbc.co.uk').registered_domain
        'bbc.co.uk'
        >>> extract('http://localhost:8080').registered_domain
        ''
        """
        if self.domain and self.suffix:
            return self.domain + '.' + self.suffix
        return ''

    @property
    def fqdn(self):
        """
        Returns a Fully Qualified Domain Name, if there is a proper domain/suffix.

        >>> extract('http://forums.bbc.co.uk/path/to/file').fqdn
        'forums.bbc.co.uk'
        >>> extract('http://localhost:8080').fqdn
        ''
        """
        if self.domain and self.suffix:
            # self is the namedtuple (subdomain domain suffix)
            return '.'.join(i for i in self if i)
        return ''

    @property
    def ipv4(self):
        """
        Returns the ipv4 if that is what the presented domain/url is

        >>> extract('http://127.0.0.1/path/to/file').ipv4
        '127.0.0.1'
        >>> extract('http://127.0.0.1.1/path/to/file').ipv4
        ''
        >>> extract('http://256.1.1.1').ipv4
        ''
        """
        if not (self.suffix or self.subdomain) and IP_RE.match(self.domain):
            return self.domain
        return ''


class TLDExtract(object):
    '''A callable for extracting, subdomain, domain, and suffix components from
    a URL.'''

    # TODO: Agreed with Pylint: too-many-arguments
    def __init__(self, cache_file=CACHE_FILE, suffix_list_urls=PUBLIC_SUFFIX_LIST_URLS,  # pylint: disable=too-many-arguments
                 fallback_to_snapshot=True, include_psl_private_domains=False, extra_suffixes=(),
                 cache_fetch_timeout=CACHE_TIMEOUT):
        """
        Constructs a callable for extracting subdomain, domain, and suffix
        components from a URL.

        Upon calling it, it first checks for a JSON `cache_file`.
        By default, the `cache_file` will live in the tldextract directory.

        You can disable the caching functionality of this module  by setting `cache_file` to False.

        If the `cache_file` does not exist (such as on the first run), HTTP request the URLs in
        `suffix_list_urls` in order, until one returns public suffix list data. To disable HTTP
        requests, set this to something falsy.

        The default list of URLs point to the latest version of the Mozilla Public Suffix List and
        its mirror, but any similar document could be specified.

        Local files can be specified by using the `file://` protocol. (See `urllib2` documentation.)

        If there is no `cache_file` loaded and no data is found from the `suffix_list_urls`,
        the module will fall back to the included TLD set snapshot. If you do not want
        this behavior, you may set `fallback_to_snapshot` to False, and an exception will be
        raised instead.

        The Public Suffix List includes a list of "private domains" as TLDs,
        such as blogspot.com. These do not fit `tldextract`'s definition of a
        suffix, so these domains are excluded by default. If you'd like them
        included instead, set `include_psl_private_domains` to True.

        You can pass additional suffixes in `extra_suffixes` argument without changing list URL

        cache_fetch_timeout is passed unmodified to the underlying request object
        per the requests documentation here:
        http://docs.python-requests.org/en/master/user/advanced/#timeouts

        cache_fetch_timeout can also be set to a single value with the
        environment variable TLDEXTRACT_CACHE_TIMEOUT, like so:

        TLDEXTRACT_CACHE_TIMEOUT="1.2"

        When set this way, the same timeout value will be used for both connect
        and read timeouts
        """
        suffix_list_urls = suffix_list_urls or ()
        self.suffix_list_urls = tuple(url.strip() for url in suffix_list_urls if url.strip())

        self.cache_file = os.path.expanduser(cache_file or '')
        self.fallback_to_snapshot = fallback_to_snapshot
        if not (self.suffix_list_urls or self.cache_file or self.fallback_to_snapshot):
            raise ValueError("The arguments you have provided disable all ways for tldextract "
                             "to obtain data. Please provide a suffix list data, a cache_file, "
                             "or set `fallback_to_snapshot` to `True`.")

        self.include_psl_private_domains = include_psl_private_domains
        self.extra_suffixes = extra_suffixes
        self._extractor = None

        self.cache_fetch_timeout = cache_fetch_timeout
        if isinstance(self.cache_fetch_timeout, STRING_TYPE):
            self.cache_fetch_timeout = float(self.cache_fetch_timeout)

    def __call__(self, url):
        """
        Takes a string URL and splits it into its subdomain, domain, and
        suffix (effective TLD, gTLD, ccTLD, etc.) component.

        >>> extract = TLDExtract()
        >>> extract('http://forums.news.cnn.com/')
        ExtractResult(subdomain='forums.news', domain='cnn', suffix='com')
        >>> extract('http://forums.bbc.co.uk/')
        ExtractResult(subdomain='forums', domain='bbc', suffix='co.uk')
        """
        netloc = SCHEME_RE.sub("", url) \
            .partition("/")[0] \
            .partition("?")[0] \
            .partition("#")[0] \
            .split("@")[-1] \
            .partition(":")[0] \
            .strip() \
            .rstrip(".")

        labels = netloc.split(".")

        def decode_punycode(label):
            if PUNY_RE.match(label):
                try:
                    return idna.decode(label.encode('ascii'))
                except UnicodeError:
                    pass
            return label

        translations = [decode_punycode(label).lower() for label in labels]
        suffix_index = self._get_tld_extractor().suffix_index(translations)

        registered_domain = ".".join(labels[:suffix_index])
        suffix = ".".join(labels[suffix_index:])

        if not suffix and netloc and looks_like_ip(netloc):
            return ExtractResult('', netloc, '')

        subdomain, _, domain = registered_domain.rpartition('.')
        return ExtractResult(subdomain, domain, suffix)

    def update(self, fetch_now=False):
        if os.path.exists(self.cache_file):
            os.unlink(self.cache_file)
        self._extractor = None
        if fetch_now:
            self._get_tld_extractor()

    @property
    def tlds(self):
        return self._get_tld_extractor().tlds

    def _get_tld_extractor(self):
        '''Get or compute this object's TLDExtractor. Looks up the TLDExtractor
        in roughly the following order, based on the settings passed to
        __init__:

        1. Memoized on `self`
        2. Local system cache file
        3. Remote PSL, over HTTP
        4. Bundled PSL snapshot file'''
        if self._extractor:
            return self._extractor

        tlds = self._get_cached_tlds()
        if tlds:
            tlds.extend(self.extra_suffixes)
            self._extractor = _PublicSuffixListTLDExtractor(tlds)
            return self._extractor
        elif self.suffix_list_urls:
            raw_suffix_list_data = find_first_response(
                self.suffix_list_urls,
                self.cache_fetch_timeout
            )
            tlds = get_tlds_from_raw_suffix_list_data(
                raw_suffix_list_data,
                self.include_psl_private_domains
            )

        if not tlds and self.fallback_to_snapshot:
            tlds = self._get_snapshot_tld_extractor()
            tlds.extend(self.extra_suffixes)
            self._extractor = _PublicSuffixListTLDExtractor(tlds)
            return self._extractor
        elif not tlds:
            raise Exception("tlds is empty, but fallback_to_snapshot is set"
                            " to false. Cannot proceed without tlds.")

        self._cache_tlds(tlds)

        tlds.extend(self.extra_suffixes)
        self._extractor = _PublicSuffixListTLDExtractor(tlds)
        return self._extractor

    def _get_cached_tlds(self):
        '''Read the local TLD cache file. Returns None on IOError or other
        error, or if this object is not set to use the cache
        file.'''
        if not self.cache_file:
            return

        try:
            with open(self.cache_file) as cache_file:
                try:
                    return json.loads(cache_file.read())
                except (IOError, ValueError) as exc:
                    LOG.error(
                        "error reading TLD cache file %s: %s",
                        self.cache_file,
                        exc
                    )
        except IOError as ioe:
            file_not_found = ioe.errno == errno.ENOENT
            if not file_not_found:
                LOG.error("error reading TLD cache file %s: %s", self.cache_file, ioe)

    @staticmethod
    def _get_snapshot_tld_extractor():
        snapshot_stream = pkg_resources.resource_stream(__name__, '.tld_set_snapshot')
        with closing(snapshot_stream) as snapshot_file:
            return json.loads(snapshot_file.read().decode('utf-8'))

    def _cache_tlds(self, tlds):
        '''Logs a diff of the new TLDs and caches them on disk, according to
        settings passed to __init__.'''
        if LOG.isEnabledFor(logging.DEBUG):
            import difflib
            snapshot_stream = pkg_resources.resource_stream(__name__, '.tld_set_snapshot')
            with closing(snapshot_stream) as snapshot_file:
                snapshot = sorted(
                    json.loads(snapshot_file.read().decode('utf-8'))
                )
            new = sorted(tlds)
            LOG.debug('computed TLD diff:\n' + '\n'.join(difflib.unified_diff(
                snapshot,
                new,
                fromfile=".tld_set_snapshot",
                tofile=self.cache_file
            )))

        if self.cache_file:
            try:
                with open(self.cache_file, 'w') as cache_file:
                    json.dump(tlds, cache_file)
            except IOError as ioe:
                LOG.warning("unable to cache TLDs in file %s: %s", self.cache_file, ioe)


TLD_EXTRACTOR = TLDExtract()


@wraps(TLD_EXTRACTOR.__call__)
def extract(url):
    return TLD_EXTRACTOR(url)


@wraps(TLD_EXTRACTOR.update)
def update(*args, **kwargs):
    return TLD_EXTRACTOR.update(*args, **kwargs)


def get_tlds_from_raw_suffix_list_data(suffix_list_source, include_psl_private_domains=False):
    if include_psl_private_domains:
        text = suffix_list_source
    else:
        text, _, _ = suffix_list_source.partition('// ===BEGIN PRIVATE DOMAINS===')

    tlds = [m.group('suffix') for m in PUBLIC_SUFFIX_RE.finditer(text)]
    return tlds


class _PublicSuffixListTLDExtractor(object):

    def __init__(self, tlds):
        self.tlds = frozenset(tlds)

    def suffix_index(self, lower_spl):
        """Returns the index of the first suffix label.
        Returns len(spl) if no suffix is found
        """
        for i in range(len(lower_spl)):
            maybe_tld = '.'.join(lower_spl[i:])
            exception_tld = '!' + maybe_tld
            if exception_tld in self.tlds:
                return i + 1

            if maybe_tld in self.tlds:
                return i

            wildcard_tld = '*.' + '.'.join(lower_spl[i + 1:])
            if wildcard_tld in self.tlds:
                return i

        return len(lower_spl)



'''tldextract integration tests.'''

import logging
import os
import traceback

import pytest

import tldextract


def test_log_snapshot_diff(mocker):
    mocker.patch.object(logging.getLogger(), 'level', logging.DEBUG)
    debug_mock = mocker.patch.object(logging.getLogger('tldextract'), 'debug')

    extractor = tldextract.TLDExtract()
    try:
        os.remove(extractor.cache_file)
    except (IOError, OSError):
        logging.warning(traceback.format_exc())

    extractor('ignore.com')

    assert debug_mock.call_count == 1
    log_str = debug_mock.call_args[0][0]
    assert log_str.startswith('computed TLD diff')


def test_bad_kwargs():
    with pytest.raises(ValueError):
        tldextract.TLDExtract(
            cache_file=False, suffix_list_urls=False, fallback_to_snapshot=False
        )



'''tldextract test helpers.'''

from subprocess import CalledProcessError, PIPE, Popen
import tempfile


def check_output(*popenargs, **kwargs):
    '''Copied from Python 2.7.'''
    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')
    process = Popen(stdout=PIPE, *popenargs, **kwargs)
    output, _ = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessError(retcode, cmd, output=output)
    return output


def temporary_file():
    """ Make a writable temporary file and return its absolute path.
    """
    return tempfile.mkstemp()[1]
